import contextlib
import logging

from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS

from pgtrigger import core, installation, registry, runtime


def _setup_logging():  # pragma: no cover
    installation.LOGGER.addHandler(logging.StreamHandler())
    if not installation.LOGGER.level:
        installation.LOGGER.setLevel(logging.INFO)


class Command(BaseCommand):
    help = "Core django-pgtrigger subcommands."

    def add_arguments(self, parser):
        subparsers = parser.add_subparsers(
            title="sub-commands",
            required=True,
        )

        ls_parser = subparsers.add_parser("ls", help="List triggers and their installation state.")
        ls_parser.add_argument("uris", nargs="*", type=str)
        ls_parser.add_argument("-d", "--database", help="The database")
        ls_parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )
        ls_parser.set_defaults(method=self.ls)

        install_parser = subparsers.add_parser("install", help="Install triggers.")
        install_parser.add_argument("uris", nargs="*", type=str)
        install_parser.add_argument("-d", "--database", help="The database")
        install_parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )
        install_parser.set_defaults(method=self.install)

        uninstall_parser = subparsers.add_parser("uninstall", help="Uninstall triggers.")
        uninstall_parser.add_argument("uris", nargs="*", type=str)
        uninstall_parser.add_argument("-d", "--database", help="The database")
        uninstall_parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )
        uninstall_parser.set_defaults(method=self.uninstall)

        enable_parser = subparsers.add_parser("enable", help="Enable triggers.")
        enable_parser.add_argument("uris", nargs="*", type=str)
        enable_parser.add_argument("-d", "--database", help="The database")
        enable_parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )
        enable_parser.set_defaults(method=self.enable)

        disable_parser = subparsers.add_parser("disable", help="Disable triggers.")
        disable_parser.add_argument("uris", nargs="*", type=str)
        disable_parser.add_argument("-d", "--database", help="The database")
        disable_parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )
        disable_parser.set_defaults(method=self.disable)

        prune_parser = subparsers.add_parser(
            "prune", help="Prune installed triggers that are no longer in the codebase."
        )
        prune_parser.add_argument("-d", "--database", help="The database")
        prune_parser.add_argument(
            "-s",
            "--schema",
            action="append",
            help="Set the search path to this schema",
        )
        prune_parser.set_defaults(method=self.prune)

    def handle(self, *args, method, **options):
        database = options["database"] or DEFAULT_DB_ALIAS
        schemas = options["schema"] or []

        if schemas:
            context = runtime.schema(*schemas, databases=[database])
        else:
            context = contextlib.nullcontext()

        with context:
            return method(*args, **options)

    def ls(self, *args, **options):
        uris = options["uris"]

        status_formatted = {
            core.UNINSTALLED: "\033[91mUNINSTALLED\033[0m",
            core.INSTALLED: "\033[92mINSTALLED\033[0m",
            core.OUTDATED: "\033[93mOUTDATED\033[0m",
            core.PRUNE: "\033[96mPRUNE\033[0m",
            core.UNALLOWED: "\033[94mUNALLOWED\033[0m",
        }

        enabled_formatted = {
            True: "\033[92mENABLED\033[0m",
            False: "\033[91mDISABLED\033[0m",
            None: "\033[94mN/A\033[0m",
        }

        def _format_status(status, enabled, uri):
            if status in (core.UNINSTALLED, core.UNALLOWED):
                enabled = None

            return status_formatted[status], enabled_formatted[enabled], uri

        formatted = []

        for model, trigger in registry.registered(*uris):
            uri = trigger.get_uri(model)
            status, enabled = trigger.get_installation_status(model, database=options["database"])
            formatted.append(_format_status(status, enabled, uri))

        if not uris:
            for trigger in installation.prunable(database=options["database"]):
                formatted.append(_format_status("PRUNE", trigger[2], f"{trigger[0]}:{trigger[1]}"))

        max_status_len = max(len(val) for val, _, _ in formatted)
        max_enabled_len = max(len(val) for _, val, _ in formatted)
        for status, enabled, uri in formatted:
            print(
                f"{{: <{max_status_len}}} {{: <{max_enabled_len}}} {{}}".format(
                    status, enabled, uri
                )
            )

    def install(self, *args, **options):
        _setup_logging()
        installation.install(*options["uris"], database=options["database"])

    def uninstall(self, *args, **options):
        _setup_logging()
        installation.uninstall(*options["uris"], database=options["database"])

    def enable(self, *args, **options):
        _setup_logging()
        installation.enable(*options["uris"], database=options["database"])

    def disable(self, *args, **options):
        _setup_logging()
        installation.disable(*options["uris"], database=options["database"])

    def prune(self, *args, **options):
        _setup_logging()
        installation.prune(database=options["database"])



import django

from pgtrigger.contrib import (
    FSM,
    Composer,
    Protect,
    ReadOnly,
    SoftDelete,
    UpdateSearchVector,
)
from pgtrigger.core import (
    After,
    AllChange,
    AllDontChange,
    AnyChange,
    AnyDontChange,
    Before,
    Condition,
    Deferred,
    Delete,
    F,
    Func,
    Immediate,
    Insert,
    InsteadOf,
    IsDistinctFrom,
    IsNotDistinctFrom,
    Level,
    Operation,
    Operations,
    Q,
    Referencing,
    Row,
    Statement,
    Timing,
    Trigger,
    Truncate,
    Update,
    UpdateOf,
    When,
)
from pgtrigger.installation import (
    disable,
    enable,
    install,
    prunable,
    prune,
    uninstall,
)
from pgtrigger.registry import (
    register,
    registered,
)
from pgtrigger.runtime import (
    constraints,
    ignore,
    is_ignored,
    schema,
)
from pgtrigger.version import __version__

if django.VERSION < (3, 2):  # pragma: no cover
    default_app_config = "pgtrigger.apps.PGTriggerConfig"

del django


__all__ = [
    "After",
    "AllChange",
    "AllDontChange",
    "AnyChange",
    "AnyDontChange",
    "Before",
    "Composer",
    "Condition",
    "constraints",
    "Deferred",
    "Delete",
    "disable",
    "enable",
    "F",
    "FSM",
    "Func",
    "ignore",
    "is_ignored",
    "Immediate",
    "Insert",
    "install",
    "InsteadOf",
    "IsDistinctFrom",
    "IsNotDistinctFrom",
    "Level",
    "Operation",
    "Operations",
    "Protect",
    "prunable",
    "prune",
    "Q",
    "ReadOnly",
    "Referencing",
    "register",
    "registered",
    "Row",
    "schema",
    "SoftDelete",
    "Statement",
    "Timing",
    "Trigger",
    "Truncate",
    "uninstall",
    "Update",
    "UpdateOf",
    "UpdateSearchVector",
    "When",
    "__version__",
]



# Generated by Django 2.2 on 2022-08-10 19:13

import django.contrib.postgres.search
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0007_auto_20220808_1055"),
    ]

    operations = [
        migrations.CreateModel(
            name="SearchModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("body_vector", django.contrib.postgres.search.SearchVectorField()),
                ("title_body_vector", django.contrib.postgres.search.SearchVectorField()),
                ("title", models.CharField(max_length=128)),
                ("body", models.TextField()),
            ],
        ),
    ]



import pytest

import pgtrigger
from pgtrigger import registry
from pgtrigger.tests import models


def test_registered_invalid_args():
    with pytest.raises(ValueError):
        pgtrigger.registered("uri")


def test_registry():
    """
    Tests dynamically registering and unregistering triggers
    """
    init_registry_size = len(registry._registry)
    # The trigger registry should already be populated with our test triggers
    assert init_registry_size >= 6

    # Add a trigger to the registry
    trigger = pgtrigger.Trigger(
        when=pgtrigger.Before,
        name="my_aliased_trigger",
        operation=pgtrigger.Insert | pgtrigger.Update,
        func="RAISE EXCEPTION 'no no no!';",
    )

    # Register/unregister in context managers. The state should be the same
    # at the end as the beginning
    with trigger.register(models.TestModel):
        assert len(registry._registry) == init_registry_size + 1
        assert f"tests.TestModel:{trigger.name}" in registry._registry

        with trigger.unregister(models.TestModel):
            assert len(registry._registry) == init_registry_size
            assert f"tests.TestModel:{trigger.name}" not in registry._registry

        # Try obtaining trigger by alias
        assert pgtrigger.registered("tests.TestModel:my_aliased_trigger")

    assert len(registry._registry) == init_registry_size
    assert f"tests.TestModel:{trigger.name}" not in registry._registry
    with pytest.raises(KeyError, match="not found"):
        pgtrigger.registered(f"tests.TestModel:{trigger.name}")

    with pytest.raises(ValueError, match="must be in the format"):
        pgtrigger.registered("tests.TestMode")


def test_duplicate_trigger_names(mocker):
    """Ensure that duplicate trigger names are properly detected"""

    # Add a trigger to the registry
    trigger1 = pgtrigger.Trigger(
        name="mytrigger", when=pgtrigger.Before, operation=pgtrigger.Insert
    )
    trigger2 = pgtrigger.Protect(
        name="mytrigger", when=pgtrigger.Before, operation=pgtrigger.Insert
    )
    trigger3 = pgtrigger.Trigger(
        name="MyTrigger", when=pgtrigger.Before, operation=pgtrigger.Insert
    )

    assert trigger1.get_pgid(models.TestModel) == "pgtrigger_mytrigger_b34c5"
    assert trigger3.get_pgid(models.TestModel) == "pgtrigger_mytrigger_4a08f"

    # Check that a conflict cannot happen in the registry.
    # NOTE - use context managers to ensure we don't keep around
    # these registered triggers in other tests
    with trigger1.register(models.TestModel):
        with pytest.raises(KeyError, match="already used"):
            with trigger2.register(models.TestModel):
                pass

    mocker.patch.object(pgtrigger.Trigger, "get_pgid", return_value="duplicate")

    # Check that a conflict cannot happen in the generated postgres ID.
    # NOTE - use context managers to ensure we don't keep around
    # these registered triggers in other tests
    with pytest.raises(KeyError, match="already in use"):
        with trigger1.register(models.TestModel):
            pass


def test_duplicate_trigger_names_proxy_model(mocker):
    """Test that duplicate trigger names are detected when using proxy models"""

    # TestTriggerProxy registers "protect_delete" for TestTrigger.
    # If we try to register this trigger directly on TestTrigger, it should result
    # in a duplicate error
    trigger = pgtrigger.Trigger(
        name="protect_delete", when=pgtrigger.Before, operation=pgtrigger.Insert
    )
    with pytest.raises(KeyError, match="already used"):
        with trigger.register(models.TestTrigger):
            pass



import contextlib

import pytest
from django.db import DEFAULT_DB_ALIAS, connections
from django.db import transaction as db_transaction
from django.db.utils import DatabaseError


@contextlib.contextmanager
def raises_trigger_error(match=None, database=DEFAULT_DB_ALIAS, transaction=None):
    with contextlib.ExitStack() as stack:
        stack.enter_context(pytest.raises(DatabaseError, match=match))

        if transaction is None:
            transaction = connections[database].in_atomic_block

        if transaction:
            stack.enter_context(db_transaction.atomic(using=database))

        yield



import pytest
from django.core.management import call_command


@pytest.fixture(scope="session")
def django_db_setup(django_db_setup, django_db_blocker, request):
    with django_db_blocker.unblock():
        # Note - schemas for databases are made in the pre-migrate hook
        # The django test runner only runs migrations ones per unique connection string.
        # Ensure that we've migrated all of our schema-based databases here
        call_command("migrate", database="default", verbosity=request.config.option.verbose)
        call_command("migrate", database="order", verbosity=request.config.option.verbose)
        call_command("migrate", database="receipt", verbosity=request.config.option.verbose)


@pytest.fixture(autouse=True)
def disable_logging(mocker):
    mocker.patch("pgtrigger.management.commands.pgtrigger._setup_logging", autospec=True)



import datetime as dt

import ddf
import django
import pgbulk
import pytest
from django.core.exceptions import FieldDoesNotExist

import pgtrigger
from pgtrigger.tests import models, utils


def test_registered_invalid_args():
    with pytest.raises(ValueError):
        pgtrigger.registered("uri")


@pytest.mark.django_db
def test_read_only():
    """Tests the ReadOnly trigger"""
    with pytest.raises(ValueError, match="only one of"):
        pgtrigger.ReadOnly(name="uneditable", fields=["level"], exclude=["hello"])

    trigger = pgtrigger.ReadOnly(name="uneditable", fields=["level"])
    with pytest.raises(FieldDoesNotExist):
        trigger.install(models.TestModel)

    trigger = pgtrigger.ReadOnly(name="uneditable", exclude=["level"])
    with pytest.raises(FieldDoesNotExist):
        trigger.install(models.TestModel)

    trigger = pgtrigger.ReadOnly(name="uneditable")
    with trigger.install(models.TestModel):
        m = ddf.G(models.TestModel, int_field=1)
        m.save()

        with utils.raises_trigger_error(match="Cannot update rows"):
            m.int_field = 2
            m.save()

    trigger = pgtrigger.ReadOnly(name="uneditable", fields=["char_field", "float_field"])
    with trigger.install(models.TestModel):
        m = ddf.G(models.TestModel, int_field=2, char_field="a")
        m.int_field = 3
        m.save()

        with utils.raises_trigger_error(match="Cannot update rows"):
            m.char_field = "b"
            m.save()

    trigger = pgtrigger.ReadOnly(name="uneditable", exclude=["int_field"])
    with trigger.install(models.TestModel):
        m = ddf.G(models.TestModel, int_field=4, char_field="a")
        m.int_field = 5
        m.save()

        with utils.raises_trigger_error(match="Cannot update rows"):
            m.char_field = "b"
            m.save()


@pytest.mark.django_db
def test_search_model():
    """Verifies search model fields are kept up to date"""
    obj = models.SearchModel.objects.create(
        title="This is a message", body="Hello World. What a great body."
    )
    models.SearchModel.objects.create(title="Hi guys", body="Random Word. This is a good idea.")
    models.SearchModel.objects.create(
        title="Hello", body="Other words. Many great ideas come from stuff."
    )
    models.SearchModel.objects.create(title="The title", body="A short message.")

    assert models.SearchModel.objects.filter(body_vector="hello").count() == 1
    assert models.SearchModel.objects.filter(body_vector="words").count() == 2
    assert models.SearchModel.objects.filter(body_vector="world").count() == 1
    assert models.SearchModel.objects.filter(title_body_vector="message").count() == 2
    assert models.SearchModel.objects.filter(title_body_vector="idea").count() == 2
    assert models.SearchModel.objects.filter(title_body_vector="hello").count() == 2

    obj.body = "Nothing more"
    obj.save()
    assert not models.SearchModel.objects.filter(body_vector="hello").exists()
    assert models.SearchModel.objects.filter(title_body_vector="hello").count() == 1


def test_update_search_vector_args():
    """Verifies arg checking for UpdateSearchVector"""
    with pytest.raises(ValueError, match='provide "vector_field"'):
        pgtrigger.UpdateSearchVector()

    with pytest.raises(ValueError, match='provide "document_fields"'):
        pgtrigger.UpdateSearchVector(vector_field="vector_field")


def test_update_search_vector_ignore():
    """Verifies UpdateSearchVector cannot be ignored"""
    trigger = pgtrigger.UpdateSearchVector(
        name="hi", vector_field="vector_field", document_fields=["hi"]
    )
    with pytest.raises(RuntimeError, match="Cannot ignore UpdateSearchVector"):
        with trigger.ignore(models.SearchModel):
            pass


@pytest.mark.django_db
def test_soft_delete():
    """
    Verifies the SoftDelete test model has the "is_active" flag set to false
    """
    soft_delete = ddf.G(models.SoftDelete, is_active=True)
    ddf.G(models.FkToSoftDelete, ref=soft_delete)
    soft_delete.delete()

    assert not models.SoftDelete.objects.get().is_active
    assert not models.FkToSoftDelete.objects.exists()


@pytest.mark.django_db
def test_soft_delete_composite_pk():
    """
    Verifies the SoftDelete test model has the "is_active" flag set to false
    """
    if django.VERSION >= (5, 2):
        models.SoftDeleteCompositePk.objects.create(is_active=True, id_1=1, id_2=1)
        soft_delete = models.SoftDeleteCompositePk.objects.create(is_active=True, id_1=2, id_2=2)
        soft_delete.delete()

        assert models.SoftDeleteCompositePk.objects.get(id_1=1, id_2=1).is_active
        assert not models.SoftDeleteCompositePk.objects.get(id_1=2, id_2=2).is_active


@pytest.mark.django_db
def test_customer_soft_delete():
    """
    Verifies the CustomSoftDelete test model has the "custom_active" flag set
    to false
    """
    soft_delete = ddf.G(models.CustomSoftDelete, custom_active=True)
    soft_delete.delete()

    assert not models.CustomSoftDelete.objects.get().custom_active


@pytest.mark.django_db
def test_soft_delete_different_values():
    """
    Tests SoftDelete with different types of fields and values
    """
    # Make the LogEntry model a soft delete model where
    # "level" is set to "inactive"
    trigger = pgtrigger.SoftDelete(name="soft_delete", field="level", value="inactive")
    with trigger.install(models.LogEntry):
        le = ddf.G(models.LogEntry, level="active")
        le.delete()
        assert models.LogEntry.objects.get().level == "inactive"
    models.LogEntry.objects.all().delete()

    # Make the LogEntry model a soft delete model where
    # "old_field" is set to None
    trigger = pgtrigger.SoftDelete(name="soft_delete", field="old_field", value=None)
    with trigger.install(models.LogEntry):
        le = ddf.G(models.LogEntry, old_field="something")
        le.delete()
        assert models.LogEntry.objects.get().old_field is None


@pytest.mark.django_db
def test_fsm():
    """
    Verifies the FSM test model cannot make invalid transitions
    """
    fsm = ddf.G(models.FSM, transition="unpublished")
    fsm.transition = "inactive"
    with utils.raises_trigger_error(match="Invalid transition"):
        fsm.save()

    fsm.transition = "published"
    fsm.save()

    # Be sure we ignore FSM when there is no transition
    fsm.save()

    with utils.raises_trigger_error(match="Invalid transition"):
        fsm.transition = "unpublished"
        fsm.save()

    fsm.transition = "inactive"
    fsm.save()


def test_fsm_args():
    """Verifies arg checking for FSM"""
    with pytest.raises(ValueError, match='provide "field"'):
        pgtrigger.FSM()

    with pytest.raises(ValueError, match='provide "transitions"'):
        pgtrigger.FSM(field="hello")

    with pytest.raises(ValueError, match='contains separator ":"'):
        pgtrigger.FSM(field="hello", transitions=[("a", ":")])

    with pytest.raises(ValueError, match='contains separator ","'):
        pgtrigger.FSM(field="hello", separator=",", transitions=[("a", ",")])

    with pytest.raises(ValueError, match="contains quotes"):
        pgtrigger.FSM(field="hello", transitions=[("a", "b'")])

    with pytest.raises(ValueError, match="contains quotes"):
        pgtrigger.FSM(field="hello", transitions=[("a", 'b"')])

    with pytest.raises(ValueError, match="single character"):
        pgtrigger.FSM(field="hello", separator="aa", transitions=[("a", "b")])

    with pytest.raises(ValueError, match="must not have quotes"):
        pgtrigger.FSM(field="hello", separator="'", transitions=[("a", "b")])


@pytest.mark.django_db
def test_protect():
    """Verify deletion protect trigger works on test model"""
    deletion_protected_model = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()


@pytest.mark.django_db
def test_protect_statement_level_insert_update():
    """Verify insert/update protect trigger works on test model"""
    cond_protect = pgtrigger.Protect(
        name="composer_protect",
        operation=pgtrigger.Insert,
        level=pgtrigger.Statement,
        condition=pgtrigger.Q(new__int_field__gt=100),
    )

    with cond_protect.install(models.TestTrigger):
        models.TestTrigger.objects.bulk_create(
            [
                models.TestTrigger(int_field=2),
                models.TestTrigger(int_field=20),
                models.TestTrigger(int_field=30),
            ]
        )
        with utils.raises_trigger_error(match="Cannot insert rows"):
            models.TestTrigger.objects.bulk_create(
                [
                    models.TestTrigger(int_field=2),
                    models.TestTrigger(int_field=200),
                ]
            )

    cond_protect = pgtrigger.Protect(
        name="composer_protect",
        operation=pgtrigger.Update,
        level=pgtrigger.Statement,
        condition=pgtrigger.Q(new__int_field__gt=100, old__int_field__lte=100),
    )

    with cond_protect.install(models.TestTrigger):
        models.TestTrigger.objects.bulk_create(
            [
                models.TestTrigger(int_field=2),
                models.TestTrigger(int_field=20),
                models.TestTrigger(int_field=30),
            ]
        )
        models.TestTrigger.objects.update(int_field=1)
        with utils.raises_trigger_error(match="Cannot update rows"):
            models.TestTrigger.objects.update(int_field=101)


@pytest.mark.django_db
def test_protect_statement_level_delete():
    """Verify deletion protect trigger works on test model"""
    cond_protect = pgtrigger.Protect(
        name="composer_protect",
        operation=pgtrigger.Delete,
        level=pgtrigger.Statement,
        condition=pgtrigger.Q(old__int_field__gt=20),
    )

    with (
        cond_protect.install(models.TestTrigger),
        pgtrigger.ignore("tests.TestTriggerProxy:protect_delete"),
    ):
        values = models.TestTrigger.objects.bulk_create(
            [
                models.TestTrigger(int_field=2),
                models.TestTrigger(int_field=20),
                models.TestTrigger(int_field=30),
            ]
        )
        print("values", values[0].int_field)
        values[0].delete()
        with utils.raises_trigger_error(match="Cannot delete rows"):
            models.TestTrigger.objects.all().delete()


@pytest.mark.django_db
def test_readonly_statement_level():
    """Verify readonly statement protect trigger works on test model"""
    cond_protect = pgtrigger.ReadOnly(
        name="composer_protect",
        level=pgtrigger.Statement,
        fields=["int_field"],
    )

    with cond_protect.install(models.TestTrigger):
        models.TestTrigger.objects.bulk_create(
            [
                models.TestTrigger(int_field=2),
                models.TestTrigger(int_field=20),
                models.TestTrigger(int_field=30),
            ]
        )
        with utils.raises_trigger_error(match="Cannot update rows"):
            models.TestTrigger.objects.update(int_field=101)


@pytest.mark.django_db
def test_custom_db_table_protect_trigger():
    """Verify custom DB table names have successful triggers"""
    deletion_protected_model = ddf.G(models.CustomTableName)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()


@pytest.mark.django_db
def test_composer_protect():
    """Verify composer trigger with protection-like trigger."""

    # Test a simple protection trigger that loops through conditional rows
    composer_raise = pgtrigger.Composer(
        name="composer_protect",
        when=pgtrigger.After,
        operation=pgtrigger.Insert,
        level=pgtrigger.Statement,
        declare=[("val", "RECORD")],
        func=pgtrigger.Func(
            """
            FOR val IN SELECT * FROM {cond_new_values}
            LOOP
            RAISE EXCEPTION 'hit condition';
            END LOOP;
            RETURN NULL;
            """
        ),
        condition=pgtrigger.Q(new__int_field__gt=0),
    )

    with composer_raise.install(models.TestTrigger):
        ddf.G(models.TestTrigger, int_field=0)
        with utils.raises_trigger_error(match="hit condition"):
            models.TestTrigger.objects.create(int_field=2)


@pytest.mark.django_db
def test_composer_protect_no_condition():
    """Verify composer trigger with protection-like trigger."""

    # Test a simple protection trigger that loops through conditional rows
    composer_raise = pgtrigger.Composer(
        name="composer_protect",
        when=pgtrigger.After,
        operation=pgtrigger.Insert,
        level=pgtrigger.Statement,
        declare=[("val", "RECORD")],
        func=pgtrigger.Func(
            """
            FOR val IN SELECT * FROM {cond_new_values}
            LOOP
            RAISE EXCEPTION 'hit condition';
            END LOOP;
            RETURN NULL;
            """
        ),
    )

    with composer_raise.install(models.TestTrigger):
        with utils.raises_trigger_error(match="hit condition"):
            models.TestTrigger.objects.create(int_field=2)


@pytest.mark.django_db
def test_composer_protect_custom_condition():
    """Verify composer trigger with a custom protection-like condition."""

    composer_raise = pgtrigger.Composer(
        name="composer_protect_update",
        when=pgtrigger.After,
        operation=pgtrigger.Update,
        level=pgtrigger.Statement,
        declare=[("val", "RECORD")],
        func=pgtrigger.Func(
            """
            FOR val IN SELECT * FROM {cond_new_values}
            LOOP
            RAISE EXCEPTION 'hit condition';
            END LOOP;
            RETURN NULL;
            """
        ),
        condition=pgtrigger.Condition("NEW.* IS DISTINCT FROM OLD.*"),
    )

    with composer_raise.install(models.TestTrigger):
        ddf.G(models.TestTrigger, int_field=0, n=5)
        # A redundant update should not trigger
        models.TestTrigger.objects.update(int_field=0)

        with utils.raises_trigger_error(match="hit condition"):
            models.TestTrigger.objects.update(int_field=2)


@pytest.mark.django_db
def test_composer_log():
    """Verify composer trigger works on test model"""
    composer_log = pgtrigger.Composer(
        name="composer_log",
        when=pgtrigger.After,
        operation=pgtrigger.Update,
        level=pgtrigger.Statement,
        declare=[("val", "RECORD")],
        func=pgtrigger.Func(
            f"""
            INSERT INTO {models.TestTrigger._meta.db_table} (field, int_field, dt_field)
            SELECT new_values.field, new_values.int_field, old_values.dt_field
            FROM {{cond_joined_values}};
            RETURN NULL;
            """
        ),
        condition=pgtrigger.Q(new__int_field__gt=0, old__int_field__lte=100),
    )

    with composer_log.install(models.TestTrigger):
        assert models.TestTrigger.objects.count() == 0
        ddf.G(models.TestTrigger, int_field=0, field="a", dt_field=dt.datetime(2015, 1, 1), n=5)
        assert models.TestTrigger.objects.count() == 5

        # The conditional statement trigger should not fire
        models.TestTrigger.objects.update(int_field=-1)
        assert models.TestTrigger.objects.count() == 5

        # The condition will fire this time, creating 5 new rows
        models.TestTrigger.objects.update(
            int_field=101, dt_field=dt.datetime(2016, 1, 2), field="b"
        )
        assert models.TestTrigger.objects.count() == 10
        # Verify the values of the last six rows
        for row in models.TestTrigger.objects.order_by("id")[6:]:
            assert row.int_field == 101
            assert row.field == "b"
            assert row.dt_field == dt.datetime(2015, 1, 1)

        # Updating should not trigger
        models.TestTrigger.objects.update(int_field=1)
        assert models.TestTrigger.objects.count() == 10

        # Do a bulk update with different values
        test_triggers = list(models.TestTrigger.objects.order_by("id"))
        # These five should not trigger
        for val in test_triggers[:5]:
            val.int_field = 0

        pgbulk.update(models.TestTrigger, test_triggers)
        assert models.TestTrigger.objects.count() == 15


@pytest.mark.parametrize(
    "operation, expected",
    [
        (pgtrigger.Insert, pgtrigger.Referencing(new="new_values")),
        (pgtrigger.Truncate, None),
        (pgtrigger.UpdateOf("field"), None),
        (pgtrigger.UpdateOf("field") | pgtrigger.Insert, None),
        (pgtrigger.Truncate | pgtrigger.Insert, None),
        (pgtrigger.Update, pgtrigger.Referencing(old="old_values", new="new_values")),
        (pgtrigger.Update | pgtrigger.Insert, None),
        (pgtrigger.Update | pgtrigger.Delete, None),
        (pgtrigger.Update | pgtrigger.Insert | pgtrigger.Delete, None),
        (pgtrigger.Delete, pgtrigger.Referencing(old="old_values")),
    ],
)
def test_composer_referencing(operation, expected):
    """Verify Composer trigger referencing."""
    composer = pgtrigger.Composer(
        name="composer_referencing",
        level=pgtrigger.Statement,
        when=pgtrigger.After,
        operation=operation,
    )
    assert composer.referencing == expected


@pytest.mark.parametrize(
    "condition, expected",
    [
        (None, 'old_values JOIN new_values ON (old_values."id") = (new_values."id")'),
        (
            pgtrigger.Condition("NEW.* IS DISTINCT FROM OLD.*"),
            'old_values JOIN new_values ON (old_values."id") = (new_values."id") WHERE (new_values.* IS DISTINCT FROM old_values.*)',  # noqa
        ),
        (
            pgtrigger.Q(new__int_field__gt=1, old__int_field__lte=100),
            'old_values JOIN new_values ON (old_values."id") = (new_values."id") WHERE (new_values."int_field" > 1 AND old_values."int_field" <= 100)',  # noqa
        ),
    ],
)
def test_composer_get_func_template_kwargs_cond_joined_values(condition, expected):
    """Verify Composer trigger get_func_template_kwargs for the cond_joined_values property."""
    assert (
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            condition=condition,
            func=pgtrigger.Func("RETURN NULL;"),
        )
        .get_func_template_kwargs(models.TestTrigger)["cond_joined_values"]
        .strip()
        == expected
    )


@pytest.mark.parametrize(
    "condition, expected",
    [
        (None, "old_values"),
        (
            pgtrigger.Condition("NEW.* IS DISTINCT FROM OLD.*"),
            'old_values JOIN new_values ON (old_values."id") = (new_values."id") WHERE (new_values.* IS DISTINCT FROM old_values.*)',  # noqa
        ),
        (
            pgtrigger.Q(old__int_field__gt=1),
            'old_values WHERE (old_values."int_field" > 1)',
        ),
        (
            pgtrigger.Q(old__int_field__gt=1) | pgtrigger.Q(new__int_field__lte=100),
            'old_values JOIN new_values ON (old_values."id") = (new_values."id") WHERE (old_values."int_field" > 1 OR new_values."int_field" <= 100)',  # noqa
        ),
    ],
)
def test_composer_get_func_template_kwargs_cond_old_values(condition, expected):
    """Verify Composer trigger get_func_template_kwargs for the cond_old_values property."""
    assert (
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            condition=condition,
            func=pgtrigger.Func("RETURN NULL;"),
        )
        .get_func_template_kwargs(models.TestTrigger)["cond_old_values"]
        .strip()
        == expected
    )


@pytest.mark.parametrize(
    "condition, expected",
    [
        (None, "new_values"),
        (
            pgtrigger.Condition("NEW.* IS DISTINCT FROM OLD.*"),
            'old_values JOIN new_values ON (old_values."id") = (new_values."id") WHERE (new_values.* IS DISTINCT FROM old_values.*)',  # noqa
        ),
        (
            pgtrigger.Q(new__int_field__gt=1),
            'new_values WHERE (new_values."int_field" > 1)',
        ),
        (
            pgtrigger.Q(old__int_field__gt=1),
            'old_values JOIN new_values ON (old_values."id") = (new_values."id") WHERE (old_values."int_field" > 1)',  # noqa
        ),
        (
            pgtrigger.Q(old__int_field__gt=1) | pgtrigger.Q(new__int_field__lte=100),
            'old_values JOIN new_values ON (old_values."id") = (new_values."id") WHERE (old_values."int_field" > 1 OR new_values."int_field" <= 100)',  # noqa
        ),
    ],
)
def test_composer_get_func_template_kwargs_cond_new_values(condition, expected):
    """Verify Composer trigger get_func_template_kwargs for the cond_new_values property."""
    assert (
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            condition=condition,
            func=pgtrigger.Func("RETURN NULL;"),
        )
        .get_func_template_kwargs(models.TestTrigger)["cond_new_values"]
        .strip()
        == expected
    )


def test_composer_properties():
    """Verify Composer trigger properties."""
    with pytest.raises(ValueError, match="referencing"):
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            referencing=pgtrigger.Referencing(new="new_values"),
        )

    func = pgtrigger.Func("RETURN NULL;")
    assert (
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            func={
                pgtrigger.Statement: func,
            },
        ).get_func(models.TestTrigger)
        == func
    )
    assert (
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Row,
            when=pgtrigger.After,
            operation=pgtrigger.Update,
            func={
                pgtrigger.Row: func,
            },
        ).get_func(models.TestTrigger)
        == func
    )

    # Verify we can't render the func if it references a non-existent transition table
    with pytest.raises(ValueError, match="references OLD"):
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Insert,
            func="SELECT * FROM old_values.*",
        ).render_func(models.TestTrigger)

    with pytest.raises(ValueError, match="references OLD"):
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update | pgtrigger.Insert,
            func="SELECT * FROM old_values.*",
        ).render_func(models.TestTrigger)

    with pytest.raises(ValueError, match="references NEW"):
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update | pgtrigger.Insert,
            func=pgtrigger.Func("SELECT * FROM {cond_old_values}"),
            condition=pgtrigger.AnyChange(),
        ).render_func(models.TestTrigger)

    with pytest.raises(ValueError, match="references NEW"):
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Delete,
            func="SELECT * FROM new_values.*",
        ).render_func(models.TestTrigger)

    with pytest.raises(ValueError, match="references NEW"):
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update | pgtrigger.Delete,
            func="SELECT * FROM new_values.*",
        ).render_func(models.TestTrigger)

    with pytest.raises(ValueError, match="references NEW"):
        pgtrigger.Composer(
            name="composer_values_properties",
            level=pgtrigger.Statement,
            when=pgtrigger.After,
            operation=pgtrigger.Update | pgtrigger.Delete,
            func=pgtrigger.Func("SELECT * FROM {cond_new_values}"),
            condition=pgtrigger.AnyChange(),
        ).render_func(models.TestTrigger)



# Generated by Django 5.2 on 2025-04-25 20:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0014_softdeletecompositepk"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="ConcreteChild",
            fields=[
                (
                    "changedcondition_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="tests.changedcondition",
                    ),
                ),
                ("child_field", models.CharField(max_length=16)),
            ],
            bases=("tests.changedcondition",),
        ),
        migrations.CreateModel(
            name="AbstractChild",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("field", models.CharField(max_length=16)),
                ("int_field", models.IntegerField(default=0)),
                ("dt_field", models.DateTimeField(auto_now=True)),
                ("nullable", models.CharField(default=None, max_length=16, null=True)),
                ("child_field", models.CharField(max_length=16)),
                (
                    "char_pk_fk_field",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="tests.charpk",
                    ),
                ),
                (
                    "fk_field",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "m2m_field",
                    models.ManyToManyField(related_name="+", to=settings.AUTH_USER_MODEL),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



"""
Functions for runtime-configuration of triggers, such as ignoring
them or dynamically setting the search path.
"""

from __future__ import annotations

import contextlib
import threading
from collections.abc import Generator
from typing import TYPE_CHECKING, List, Union

from django.db import connections

from pgtrigger import registry, utils

if utils.psycopg_maj_version == 2:
    import psycopg2.extensions
    import psycopg2.sql as psycopg_sql
elif utils.psycopg_maj_version == 3:
    import psycopg.pq
    import psycopg.sql as psycopg_sql
else:
    raise AssertionError

if TYPE_CHECKING:
    from django.db.backends.utils import CursorWrapper
    from typing_extensions import TypeAlias

    from pgtrigger import Timing

_Query: "TypeAlias" = "str | bytes | psycopg_sql.SQL | psycopg_sql.Composed"

# All triggers currently being ignored
_ignore = threading.local()

# All schemas in the search path
_schema = threading.local()


def _query_to_str(query: _Query, cursor: CursorWrapper) -> str:
    if isinstance(query, str):
        return query
    elif isinstance(query, bytes):
        return query.decode()
    elif isinstance(query, (psycopg_sql.SQL, psycopg_sql.Composed)):
        return query.as_string(cursor.connection)
    else:  # pragma: no cover
        raise TypeError(f"Unsupported query type: {type(query)}")


def _is_concurrent_statement(sql: _Query, cursor: CursorWrapper) -> bool:
    """
    True if the sql statement is concurrent and cannot be ran in a transaction
    """
    sql = _query_to_str(sql, cursor)
    sql = sql.strip().lower() if sql else ""
    return sql.startswith("create") and "concurrently" in sql


def _is_transaction_errored(cursor):
    """
    True if the current transaction is in an errored state
    """
    if utils.psycopg_maj_version == 2:
        return (
            cursor.connection.get_transaction_status()
            == psycopg2.extensions.TRANSACTION_STATUS_INERROR
        )
    elif utils.psycopg_maj_version == 3:
        return cursor.connection.info.transaction_status == psycopg.pq.TransactionStatus.INERROR
    else:
        raise AssertionError


def _can_inject_variable(cursor, sql):
    """True if we can inject a SQL variable into a statement.

    A named cursor automatically prepends
    "NO SCROLL CURSOR WITHOUT HOLD FOR" to the query, which
    causes invalid SQL to be generated. There is no way
    to override this behavior in psycopg, so ignoring triggers
    cannot happen for named cursors. Django only names cursors
    for iterators and other statements that read the database,
    so it seems to be safe to ignore named cursors.

    Concurrent index creation is also incompatible with local variable
    setting. Ignore these cases for now.
    """
    return (
        not getattr(cursor, "name", None)
        and not _is_concurrent_statement(sql, cursor)
        and not _is_transaction_errored(cursor)
    )


def _execute_wrapper(execute_result):
    if utils.psycopg_maj_version == 3:
        while execute_result is not None and execute_result.nextset():
            pass
    return execute_result


def _inject_pgtrigger_ignore(execute, sql, params, many, context):
    """
    A connection execution wrapper that sets a pgtrigger.ignore
    variable in the executed SQL. This lets other triggers know when
    they should ignore execution
    """
    if _can_inject_variable(context["cursor"], sql):
        serialized_ignore = "{" + ",".join(_ignore.value) + "}"
        sql = _query_to_str(sql, context["cursor"])
        sql = f"SELECT set_config('pgtrigger.ignore', %s, true); {sql}"
        params = [serialized_ignore, *(params or ())]

    return _execute_wrapper(execute(sql, params, many, context))


@contextlib.contextmanager
def _set_ignore_session_state(database=None):
    """Starts a session where triggers can be ignored"""
    connection = utils.connection(database)
    if _inject_pgtrigger_ignore not in connection.execute_wrappers:
        with connection.execute_wrapper(_inject_pgtrigger_ignore):
            try:
                yield
            finally:
                if connection.in_atomic_block:
                    # We've finished ignoring triggers and are in a transaction,
                    # so flush the local variable.
                    with connection.cursor() as cursor:
                        cursor.execute("SELECT set_config('pgtrigger.ignore', NULL, false);")
    else:
        yield


@contextlib.contextmanager
def _ignore_session(databases=None):
    """Starts a session where triggers can be ignored"""
    with contextlib.ExitStack() as stack:
        for database in utils.postgres_databases(databases):
            stack.enter_context(_set_ignore_session_state(database=database))

        yield


@contextlib.contextmanager
def _set_ignore_state(model, trigger):
    """
    Manage state to ignore a single URI
    """
    if not hasattr(_ignore, "value"):
        _ignore.value = set()

    pgid = trigger.get_pgid(model)
    if pgid not in _ignore.value:
        # In order to preserve backwards compatibiliy with older installations
        # of the _pgtrigger_ignore func, we must set a full URI (old version)
        # and trigger ID (new version).
        # This will be removed in version 5
        uri = f"{model._meta.db_table}:{pgid}"

        try:
            _ignore.value.add(uri)
            _ignore.value.add(pgid)
            yield
        finally:
            _ignore.value.remove(uri)
            _ignore.value.remove(pgid)
    else:  # The trigger is already being ignored
        yield


@contextlib.contextmanager
def ignore(*uris: str, databases: Union[List[str], None] = None) -> Generator[None]:
    """
    Dynamically ignore registered triggers matching URIs from executing in
    an individual thread.
    If no URIs are provided, ignore all pgtriggers from executing in an
    individual thread.

    Args:
        *uris: Trigger URIs to ignore. If none are provided, all
            triggers will be ignored.
        databases: The databases to use. If none, all postgres databases
            will be used.

    Example:
        Ingore triggers in a context manager:

            with pgtrigger.ignore("my_app.Model:trigger_name"):
                # Do stuff while ignoring trigger

    Example:
        Ignore multiple triggers as a decorator:

            @pgtrigger.ignore("my_app.Model:trigger_name", "my_app.Model:other_trigger")
            def my_func():
                # Do stuff while ignoring trigger
    """
    with contextlib.ExitStack() as stack:
        stack.enter_context(_ignore_session(databases=databases))

        for model, trigger in registry.registered(*uris):
            stack.enter_context(_set_ignore_state(model, trigger))

        yield


ignore.session = _ignore_session


def is_ignored(uri: str) -> bool:
    """
    Check if a trigger URI is currently being ignored.

    Args:
        uri: The trigger URI to check.

    Example:
        Check if a trigger is being ignored:

            with pgtrigger.ignore("my_app.Model:trigger_name"):
                pgtrigger.is_ignored("my_app.Model:trigger_name")  # Returns True

            pgtrigger.is_ignored("my_app.Model:trigger_name")  # Returns False
    """
    model, trigger = registry.registered(uri)[0]
    pg_trigger_id = trigger.get_pgid(model)
    ignored_pg_trigger_ids = getattr(_ignore, "value", set())
    return pg_trigger_id in ignored_pg_trigger_ids


def _inject_schema(execute, sql, params, many, context):
    """
    A connection execution wrapper that sets the schema
    variable in the executed SQL.
    """
    if _can_inject_variable(context["cursor"], sql) and _schema.value:
        path = ", ".join(val if not val.startswith("$") else f'"{val}"' for val in _schema.value)
        sql = f"SELECT set_config('search_path', %s, true); {sql}"
        params = [path, *(params or ())]

    return _execute_wrapper(execute(sql, params, many, context))


@contextlib.contextmanager
def _set_schema_session_state(database=None):
    connection = utils.connection(database)

    if _inject_schema not in connection.execute_wrappers:
        if connection.in_atomic_block:
            # If this is the first time we are setting the search path,
            # register the pre_execute_hook and store a reference to the original
            # search path. Note that we must use this approach because we cannot
            # simply reset the search_path at the end. A user may have previously
            # set it
            with connection.cursor() as cursor:
                cursor.execute("SELECT current_setting('search_path')")
                initial_search_path = cursor.fetchall()[0][0]

        with connection.execute_wrapper(_inject_schema):
            try:
                yield
            finally:
                if connection.in_atomic_block:
                    # We've finished modifying the search path and are in a transaction,
                    # so flush the local variable
                    with connection.cursor() as cursor:
                        cursor.execute(
                            "SELECT set_config('search_path', %s, false)", [initial_search_path]
                        )
    else:
        yield


@contextlib.contextmanager
def _schema_session(databases=None):
    """Starts a session where the search path can be modified"""
    with contextlib.ExitStack() as stack:
        for database in utils.postgres_databases(databases):
            stack.enter_context(_set_schema_session_state(database=database))

        yield


@contextlib.contextmanager
def _set_schema_state(*schemas):
    if not hasattr(_schema, "value"):
        # Use a list instead of a set because ordering is important to the search path
        _schema.value = []

    schemas = [s for s in schemas if s not in _schema.value]
    try:
        _schema.value.extend(schemas)
        yield
    finally:
        for s in schemas:
            _schema.value.remove(s)


@contextlib.contextmanager
def schema(*schemas: str, databases: Union[List[str], None] = None) -> Generator[None]:
    """
    Sets the search path to the provided schemas.

    If nested, appends the schemas to the search path if not already in it.

    Args:
        *schemas: Schemas that should be appended to the search path.
            Schemas already in the search path from nested calls will not be
            appended.
        databases: The databases to set the search path. If none, all postgres
            databases will be used.
    """
    with contextlib.ExitStack() as stack:
        stack.enter_context(_schema_session(databases=databases))
        stack.enter_context(_set_schema_state(*schemas))

        yield


schema.session = _schema_session


def constraints(timing: "Timing", *uris: str, databases: Union[List[str], None] = None) -> None:
    """
    Set deferrable constraint timing for the given triggers, which
    will persist until overridden or until end of transaction.
    Must be in a transaction to run this.

    Args:
        timing: The timing value that overrides the default trigger timing.
        *uris: Trigger URIs over which to set constraint timing.
            If none are provided, all trigger constraint timing will
            be set. All triggers must be deferrable.
        databases: The databases on which to set constraints. If none, all
            postgres databases will be used.

    Raises:
        RuntimeError: If the database of any triggers is not in a transaction.
        ValueError: If any triggers are not deferrable.
    """

    for model, trigger in registry.registered(*uris):
        if not trigger.timing:
            raise ValueError(
                f"Trigger {trigger.name} on model {model._meta.label_lower} is not deferrable."
            )

    for database in utils.postgres_databases(databases):
        if not connections[database].in_atomic_block:
            raise RuntimeError(f'Database "{database}" is not in a transaction.')

        names = ", ".join(trigger.get_pgid(model) for model, trigger in registry.registered(*uris))

        with connections[database].cursor() as cursor:
            cursor.execute(f"SET CONSTRAINTS {names} {timing}")



import collections
from typing import TYPE_CHECKING, Callable, List, Tuple

from pgtrigger import features

_unset = object()


if TYPE_CHECKING:
    from django.db.models import Model

    from pgtrigger.core import Trigger


# All registered triggers for each model
class _Registry(collections.UserDict):
    @property
    def pg_function_names(self):
        """
        The postgres function names of all registered triggers
        """
        return {trigger.get_pgid(model) for model, trigger in self.values()}

    @property
    def by_db_table(self):
        """
        Return the registry keys by db_table, name
        """
        return {(model._meta.db_table, trigger.name): trigger for model, trigger in self.values()}

    def __getitem__(self, key):
        assert isinstance(key, str)
        if len(key.split(":")) == 1:
            raise ValueError(
                'Trigger URI must be in the format of "app_label.model_name:trigger_name"'
            )
        elif key not in _registry:
            raise KeyError(f'URI "{key}" not found in pgtrigger registry')

        return super().__getitem__(key)

    def __setitem__(self, key, value):
        assert isinstance(key, str)
        model, trigger = value
        assert f"{model._meta.label}:{trigger.name}" == key

        found_trigger = self.by_db_table.get((model._meta.db_table, trigger.name))

        if not found_trigger or found_trigger != trigger:
            if found_trigger:
                raise KeyError(
                    f'Trigger name "{trigger.name}" already'
                    f' used for model "{model._meta.label}"'
                    f' table "{model._meta.db_table}".'
                )

            if trigger.get_pgid(model) in self.pg_function_names:
                raise KeyError(
                    f'Trigger "{trigger.name}" on model "{model._meta.label}"'
                    " has Postgres function name that's already in use."
                    " Use a different name for the trigger."
                )

        # Add the trigger to Meta.triggers.
        # Note, pgtrigger's App.ready() method auto-registers any
        # triggers in Meta already, meaning the trigger may already exist. If so, ignore it
        if features.migrations():  # pragma: no branch
            if trigger not in getattr(model._meta, "triggers", []):
                model._meta.triggers = list(getattr(model._meta, "triggers", [])) + [trigger]

            if trigger not in model._meta.original_attrs.get("triggers", []):
                model._meta.original_attrs["triggers"] = list(
                    model._meta.original_attrs.get("triggers", [])
                ) + [trigger]

        return super().__setitem__(key, value)

    def __delitem__(self, key):
        model, trigger = self[key]

        super().__delitem__(key)

        # If we support migration integration, remove from Meta triggers
        if features.migrations():  # pragma: no branch
            model._meta.triggers.remove(trigger)
            # If model._meta.triggers and the original_attrs triggers are the same,
            # we don't need to remove it from the original_attrs
            if trigger in model._meta.original_attrs["triggers"]:  # pragma: no branch
                model._meta.original_attrs["triggers"].remove(trigger)


_registry = _Registry()


def set(uri: str, *, model: "Model", trigger: "Trigger") -> None:
    """Set a trigger in the registry

    Args:
        uri: The trigger URI
        model: The trigger model
        trigger: The trigger object
    """
    _registry[uri] = (model, trigger)


def delete(uri: str) -> None:
    """Delete a trigger from the registry.

    Args:
        uri: The trigger URI
    """
    del _registry[uri]


def registered(*uris: str) -> List[Tuple["Model", "Trigger"]]:
    """
    Get registered trigger objects.

    Args:
        *uris: URIs of triggers to get. If none are provided,
            all triggers are returned. URIs are in the format of
            `{app_label}.{model_name}:{trigger_name}`.

    Returns:
        Matching trigger objects.
    """
    uris = uris or _registry.keys()
    return [_registry[uri] for uri in uris]


def register(*triggers: "Trigger") -> Callable:
    """
    Register the given triggers with wrapped Model class.

    Args:
        *triggers: Trigger classes to register.

    Example:
        Register by decorating a model:

            @pgtrigger.register(
                pgtrigger.Protect(
                    name="append_only",
                    operation=(pgtrigger.Update | pgtrigger.Delete)
                )
            )
            class MyModel(models.Model):
                pass

    Example:
        Register by calling functionally:

            pgtrigger.register(trigger_object)(MyModel)
    """

    def _model_wrapper(model_class):
        for trigger in triggers:
            trigger.register(model_class)

        return model_class

    return _model_wrapper



# Generated by Django 3.2.15 on 2022-08-17 22:11

import psqlextra.backend.migrations.operations.add_default_partition
import psqlextra.backend.migrations.operations.create_partitioned_model
import psqlextra.manager.manager
import psqlextra.models.partitioned
import psqlextra.types
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0009_orderschema_receiptschema"),
    ]

    operations = [
        psqlextra.backend.migrations.operations.create_partitioned_model.PostgresCreatePartitionedModel(  # noqa
            name="PartitionModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("name", models.TextField()),
                ("timestamp", models.DateTimeField()),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            partitioning_options={
                "method": psqlextra.types.PostgresPartitioningMethod["RANGE"],
                "key": ["timestamp"],
            },
            bases=(psqlextra.models.partitioned.PostgresPartitionedModel,),
            managers=[
                ("objects", psqlextra.manager.manager.PostgresManager()),
            ],
        ),
        psqlextra.backend.migrations.operations.add_default_partition.PostgresAddDefaultPartition(
            model_name="PartitionModel",
            name="default",
        ),
    ]



# Generated by Django 3.2.15 on 2022-08-18 10:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0011_auto_20220817_2211"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="partitionmodel",
            options={},
        ),
    ]



from django.apps.registry import Apps
from django.db import models
from django.utils import timezone

import pgtrigger

syncdb_apps = Apps()


class NoMigrationModel(models.Model):
    """
    For testing triggers installed with syncdb
    """

    field = models.CharField(max_length=16)
    int_field = models.IntegerField(default=0)
    dt_field = models.DateTimeField(default=timezone.now)
    nullable = models.CharField(null=True, default=None, max_length=16)

    class Meta:
        apps = syncdb_apps
        triggers = [
            pgtrigger.Trigger(
                name="protect_misc_insert",
                when=pgtrigger.Before,
                operation=pgtrigger.Insert,
                func="RAISE EXCEPTION 'no no no!';",
                condition=pgtrigger.Q(new__field="misc_insert"),
            ),
        ]



import django.apps
import django.db.backends.postgresql.schema as postgresql_schema
from django.conf import settings
from django.core.management.commands import makemigrations, migrate
from django.db.migrations import state
from django.db.models import options
from django.db.models.signals import post_migrate
from django.db.utils import load_backend

from pgtrigger import core, features, installation, migrations

# Allow triggers to be specified in model Meta. Users can turn this
# off via settings if it causes issues. If turned off, migrations
# are also disabled
if features.model_meta():  # pragma: no branch
    if "triggers" not in options.DEFAULT_NAMES:  # pragma: no branch
        options.DEFAULT_NAMES = tuple(options.DEFAULT_NAMES) + ("triggers",)


def patch_migrations():
    """
    Patch the autodetector and model state detection if migrations are turned on
    """
    if features.migrations():  # pragma: no branch
        if "triggers" not in state.DEFAULT_NAMES:  # pragma: no branch
            state.DEFAULT_NAMES = tuple(state.DEFAULT_NAMES) + ("triggers",)

        if not issubclass(  # pragma: no branch
            makemigrations.MigrationAutodetector, migrations.MigrationAutodetectorMixin
        ):
            makemigrations.MigrationAutodetector = type(
                "MigrationAutodetector",
                (migrations.MigrationAutodetectorMixin, makemigrations.MigrationAutodetector),
                {},
            )

        if not issubclass(  # pragma: no branch
            migrate.MigrationAutodetector, migrations.MigrationAutodetectorMixin
        ):
            migrate.MigrationAutodetector = type(
                "MigrationAutodetector",
                (migrations.MigrationAutodetectorMixin, migrate.MigrationAutodetector),
                {},
            )

        if django.VERSION >= (5, 2):
            makemigrations.Command.autodetector = makemigrations.MigrationAutodetector
            migrate.Command.autodetector = makemigrations.MigrationAutodetector


def patch_schema_editor():
    """
    Patch the schema editor to allow for column types to be altered on
    trigger conditions
    """
    if features.schema_editor():  # pragma: no branch
        for config in settings.DATABASES.values():
            backend = load_backend(config["ENGINE"])
            schema_editor_class = backend.DatabaseWrapper.SchemaEditorClass

            if (
                schema_editor_class
                and issubclass(
                    schema_editor_class,
                    postgresql_schema.DatabaseSchemaEditor,
                )
                and not issubclass(schema_editor_class, migrations.DatabaseSchemaEditorMixin)
            ):
                backend.DatabaseWrapper.SchemaEditorClass = type(
                    "DatabaseSchemaEditor",
                    (migrations.DatabaseSchemaEditorMixin, schema_editor_class),
                    {},
                )


def register_triggers_from_meta():
    """
    Populate the trigger registry from model `Meta.triggers`
    """
    if features.model_meta():  # pragma: no branch
        for model in django.apps.apps.get_models():
            triggers = getattr(model._meta, "triggers", [])
            for trigger in triggers:
                if not isinstance(trigger, core.Trigger):  # pragma: no cover
                    raise TypeError(f"Triggers in {model} Meta must be pgtrigger.Trigger classes")

                trigger.register(model)


def install_on_migrate(using, **kwargs):
    if features.install_on_migrate():
        installation.install(database=using)


class PGTriggerConfig(django.apps.AppConfig):
    name = "pgtrigger"

    def ready(self):
        """
        Do all necessary patching, trigger setup, and signal handler configuration
        """
        patch_migrations()
        patch_schema_editor()
        register_triggers_from_meta()

        # Configure triggers to automatically be installed after migrations
        post_migrate.connect(install_on_migrate, sender=self)



import ddf
import pytest
from django.db import IntegrityError, connection, transaction

import pgtrigger
import pgtrigger.utils
from pgtrigger.tests import models, utils

if pgtrigger.utils.psycopg_maj_version == 3:
    from psycopg.sql import SQL, Identifier
elif pgtrigger.utils.psycopg_maj_version == 2:
    from psycopg2.sql import SQL, Identifier
else:
    raise AssertionError


@pytest.mark.django_db
def test_schema():
    """Verifies behavior of pgtrigger.schema"""

    def _search_path():
        with connection.cursor() as cursor:
            cursor.execute("SHOW search_path;")
            return cursor.fetchall()[0][0]

    assert _search_path() == '"$user", public'

    with pgtrigger.schema("hello"):
        assert _search_path() == "hello"

        with pgtrigger.schema("hello", "$user"):
            assert _search_path() == 'hello, "$user"'

        assert _search_path() == "hello"

    with connection.cursor() as cursor:
        cursor.execute("SET search_path=custom;")

    with transaction.atomic():
        assert _search_path() == "custom"

        with pgtrigger.schema("hello", databases=["default"]):
            assert _search_path() == "hello"

        assert _search_path() == "custom"

    with pgtrigger.schema.session(databases=["default"]):
        assert _search_path() == "custom"

    assert _search_path() == "custom"


@pytest.mark.django_db(transaction=True)
def test_constraints():
    """
    Tests running [pgtrigger.constraints][] on deferrable triggers
    """
    # Not every trigger is deferrable, so this should raise an error
    with transaction.atomic():
        with pytest.raises(ValueError, match="is not deferrable"):
            pgtrigger.constraints(pgtrigger.Immediate)

    # Make the LogEntry model a soft delete model where
    # "level" is set to "inactive"
    trigger = pgtrigger.Protect(
        name="protect_delete",
        when=pgtrigger.After,
        operation=pgtrigger.Delete,
        timing=pgtrigger.Deferred,
    )
    with trigger.register(models.TestModel), trigger.install(models.TestModel):
        # Verify we have to be in a transaction
        with pytest.raises(RuntimeError, match="not in a transaction"):
            pgtrigger.constraints(pgtrigger.Immediate, "tests.TestModel:protect_delete")

        obj = ddf.G(models.TestModel)
        with transaction.atomic():
            # This "with" is only here to validate that ignoring the trigger will
            # NOT happen. After this "with" is done, the transaction still hasn't finished
            # and the trigger hasn't executed yet, so it won't be ignored.
            with pgtrigger.ignore("tests.TestModel:protect_delete"):
                obj.delete()
                # Deletion works within the transaction so far since trigger is deferred
                assert not models.TestModel.objects.exists()

            # When we set constraints to Immediate, it should fail inside
            # of the transaction
            with utils.raises_trigger_error(match="Cannot delete", transaction=False):
                # The first statement does nothing because the trigger is already deferred
                pgtrigger.constraints(
                    pgtrigger.Deferred, "tests.TestModel:protect_delete", databases=["default"]
                )
                pgtrigger.constraints(
                    pgtrigger.Immediate, "tests.TestModel:protect_delete", databases=["default"]
                )


@pytest.mark.django_db(transaction=True)
def test_ignore_nested_transactions():
    """Verifies a trigger can be ignored during nested transactions"""
    ddf.G(models.CustomTableName, int_field=1)
    trigger = pgtrigger.Protect(
        name="protect_insert",
        when=pgtrigger.Before,
        operation=pgtrigger.Insert,
    )
    with trigger.register(models.CustomTableName):
        with trigger.install(models.CustomTableName):
            with transaction.atomic():
                with pgtrigger.ignore("tests.CustomTableName:protect_insert"):
                    try:
                        with transaction.atomic():  # pragma: no branch
                            models.CustomTableName.objects.create(int_field=1)
                    except IntegrityError:
                        models.CustomTableName.objects.create(int_field=2)


@pytest.mark.django_db(transaction=True)
def test_ignore_session():
    """Verifies an ignore session can be used to avoid transaction-related issues"""
    ddf.G(models.CustomTableName, int_field=1)
    trigger = pgtrigger.Protect(
        name="protect_insert",
        when=pgtrigger.Before,
        operation=pgtrigger.Insert,
    )
    with trigger.register(models.CustomTableName), trigger.install(models.CustomTableName):
        with pgtrigger.ignore.session():
            with transaction.atomic():
                with pgtrigger.ignore("tests.CustomTableName:protect_insert"):
                    try:
                        models.CustomTableName.objects.create(int_field=1)
                    except IntegrityError:
                        pass


@pytest.mark.django_db
def test_ignore_no_transaction_leaks():
    """Verify ignore does not leak during a transaction"""
    deletion_protected_model = ddf.G(models.TestTrigger)
    with pgtrigger.ignore("tests.TestTriggerProxy:protect_delete"):
        deletion_protected_model.delete()
        assert not models.TestTrigger.objects.exists()

    deletion_protected_model = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()


@pytest.mark.django_db
@pytest.mark.parametrize("model_class", [models.TestTriggerProxy, models.CustomTableName])
def test_basic_ignore(model_class):
    """Verify basic dynamic ignore functionality"""
    deletion_protected_model = ddf.G(model_class)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()

    with pgtrigger.ignore(f"tests.{model_class.__name__}:protect_delete"):
        deletion_protected_model.delete()

    assert not models.TestTrigger.objects.exists()

    deletion_protected_model = ddf.G(model_class)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()

    # Verify that named cursors are ignored and that valid SQL is still generated
    with pgtrigger.ignore(f"tests.{model_class.__name__}:protect_delete"):
        assert len(list(model_class.objects.all().iterator())) == 1


@pytest.mark.django_db
def test_nested_ignore():
    """Test nesting pgtrigger.ignore()"""
    deletion_protected_model1 = ddf.G(models.TestTrigger)
    deletion_protected_model2 = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model1.delete()

    with pgtrigger.ignore("tests.TestTriggerProxy:protect_delete"):
        with pgtrigger.ignore("tests.TestTriggerProxy:protect_delete"):
            deletion_protected_model1.delete()
        deletion_protected_model2.delete()

    assert not models.TestTrigger.objects.exists()

    deletion_protected_model = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()

    with pgtrigger.ignore.session(databases=["default"]):
        deletion_protected_model = ddf.G(models.TestTrigger)
        with utils.raises_trigger_error(match="Cannot delete rows"):
            deletion_protected_model.delete()


@pytest.mark.django_db
def test_multiple_ignores():
    """Tests multiple pgtrigger.ignore()"""
    deletion_protected_model1 = ddf.G(models.TestTrigger)
    ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model1.delete()

    ddf.G(models.TestTrigger, field="hi!")
    with utils.raises_trigger_error(match="no no no!"):
        models.TestTrigger.objects.create(field="misc_insert")

    with pgtrigger.ignore("tests.TestTriggerProxy:protect_delete"):
        deletion_protected_model1.delete()
        with utils.raises_trigger_error(match="no no no!"):
            models.TestTrigger.objects.create(field="misc_insert")

        with pgtrigger.ignore("tests.TestTrigger:protect_misc_insert"):
            m = models.TestTrigger.objects.create(field="misc_insert")
            m.delete()

        models.TestTrigger.objects.all().delete()

    assert not models.TestTrigger.objects.exists()

    deletion_protected_model = ddf.G(models.TestTrigger)
    with utils.raises_trigger_error(match="Cannot delete rows"):
        deletion_protected_model.delete()


@pytest.mark.django_db
def test_custom_db_table_ignore():
    """Verify we can ignore triggers on custom table names"""
    deletion_protected_model = ddf.G(models.CustomTableName)

    # Ensure we can ignore the deletion trigger
    with pgtrigger.ignore("tests.CustomTableName:protect_delete"):
        deletion_protected_model.delete()
        assert not models.CustomTableName.objects.exists()


@pytest.mark.skipif(
    pgtrigger.utils.psycopg_maj_version == 3, reason="Psycopg2 preserves entire query"
)
@pytest.mark.django_db
@pytest.mark.parametrize(
    "sql, params",
    [
        ("select count(*) from auth_user where id = %s", (1,)),
        ("select count(*) from auth_user", ()),
        (b"select count(*) from auth_user where id = %s", (1,)),
        (b"select count(*) from auth_user", ()),
    ],
)
def test_inject_trigger_ignore(settings, mocker, sql, params):
    settings.DEBUG = True
    expected_sql_base = "SELECT set_config('pgtrigger.ignore', '{ignored_triggers}', true)"
    # Order isn't deterministic, so we need to check for either order.
    expected_sql_1 = expected_sql_base.format(
        ignored_triggers=r"{tests_testtrigger:pgtrigger_protect_delete_b7483,pgtrigger_protect_delete_b7483}"
    )
    expected_sql_2 = expected_sql_base.format(
        ignored_triggers=r"{pgtrigger_protect_delete_b7483,tests_testtrigger:pgtrigger_protect_delete_b7483}"
    )
    with pgtrigger.ignore("tests.TestTriggerProxy:protect_delete"):
        with connection.cursor() as cursor:
            cursor.execute(sql, params)
            query = connection.queries[-1]
            assert query["sql"].startswith(expected_sql_1) or query["sql"].startswith(
                expected_sql_2
            )


@pytest.mark.django_db
def test_test_trigger_ignore_psycopg_sql_objects():
    """Verify that native psycopg SQL objects are handled correctly when ignoring triggers."""
    # Test with a `sql.SQL` object
    with pgtrigger.ignore("tests.TestTrigger:protect_misc_insert"), connection.cursor() as cursor:
        cursor.execute(
            SQL(
                "INSERT INTO tests_testtrigger (field, int_field, dt_field)"
                "VALUES ('misc_insert', 1, now())",
            )
        )
    # Test with a `sql.Composed` object (built through formatting)
    with pgtrigger.ignore("tests.TestTrigger:protect_misc_insert"), connection.cursor() as cursor:
        cursor.execute(
            SQL(
                "INSERT INTO {table} (field, int_field, dt_field) VALUES ('misc_insert', 1, now())"
            ).format(table=Identifier("tests_testtrigger"))
        )


@pytest.mark.django_db
def test_is_ignored():
    assert not pgtrigger.is_ignored("tests.TestTrigger:protect_misc_insert")

    with pgtrigger.ignore("tests.TestTrigger:protect_misc_insert"):
        assert pgtrigger.is_ignored("tests.TestTrigger:protect_misc_insert")

    assert not pgtrigger.is_ignored("tests.TestTrigger:protect_misc_insert")

    # Different trigger being ignored
    with pgtrigger.ignore("tests.TestTriggerProxy:protect_delete"):
        assert not pgtrigger.is_ignored("tests.TestTrigger:protect_misc_insert")

    with pytest.raises(
        ValueError,
        match='Trigger URI must be in the format of "app_label.model_name:trigger_name"',
    ):
        pgtrigger.is_ignored("wrongformat")

    with pytest.raises(KeyError, match='URI "doesnot:exist" not found in pgtrigger registry'):
        pgtrigger.is_ignored("doesnot:exist")



import collections
import hashlib
from enum import Enum
from typing import Any, Optional, Sequence, Union

import pgtrigger
from pgtrigger import utils


class _Unset(Enum):
    token = 0


_unset = _Unset.token


class UpsertTriggerSql(collections.UserString):
    """SQL for inserting or updating a trigger

    This class is intended to be versionable since migrations
    reference it. Older migrations need to be able to point
    to earlier versions of the installation template used
    for triggers.
    """

    def get_template(self) -> str:
        """
        This is v1 of the installation template. Do NOT edit
        this template unless you are absolutely sure it is
        backwards compatible, otherwise it may affect migrations
        that reference it.

        If it does need to be changed, we will need to introduce
        a version variable to be backwards compatible.

        Note: Postgres 14 has CREATE OR REPLACE syntax that
        we might consider using. This SQL is executed in
        a transaction, so dropping and recreating shouldn't
        be a problem.
        """
        return """
            CREATE OR REPLACE FUNCTION {ignore_func_name}(
                trigger_name NAME
            )
            RETURNS BOOLEAN AS $$
                DECLARE
                    _pgtrigger_ignore TEXT[];
                    _result BOOLEAN;
                BEGIN
                    BEGIN
                        SELECT INTO _pgtrigger_ignore
                            CURRENT_SETTING('pgtrigger.ignore');
                        EXCEPTION WHEN OTHERS THEN
                    END;
                    IF _pgtrigger_ignore IS NOT NULL THEN
                        SELECT trigger_name = ANY(_pgtrigger_ignore)
                        INTO _result;
                        RETURN _result;
                    ELSE
                        RETURN FALSE;
                    END IF;
                END;
            $$ LANGUAGE plpgsql;

            CREATE OR REPLACE FUNCTION {pgid}()
            RETURNS TRIGGER AS $$
                {declare}
                BEGIN
                    IF ({ignore_func_name}(TG_NAME) IS TRUE) THEN
                        IF (TG_OP = 'DELETE') THEN
                            RETURN OLD;
                        ELSE
                            RETURN NEW;
                        END IF;
                    END IF;
                    {func}
                END;
            $$ LANGUAGE plpgsql;

            DROP TRIGGER IF EXISTS {pgid} ON {table};
            CREATE {constraint} TRIGGER {pgid}
                {when} {operation} ON {table}
                {timing}
                {referencing}
                FOR EACH {level} {condition}
                EXECUTE PROCEDURE {execute};

            COMMENT ON TRIGGER {pgid} ON {table} IS '{hash}';
        """

    def get_defaults(self, pgid: str) -> dict[str, str]:
        """
        These are the default values for the installation
        template. Do NOT edit these default values. Keys
        may be added, but existing keys should never be updated,
        otherwise existing migrations may no longer be correct.

        If it does need to be changed, we will need to introduce
        a version variable to be backwards compatible.
        """
        return {
            "ignore_func_name": '"public"._pgtrigger_should_ignore',
            "declare": "",
            "constraint": "",
            "timing": "",
            "referencing": "",
            "level": "ROW",
            "condition": "",
            "execute": f"{pgid}()",
        }

    def __init__(
        self,
        *,
        ignore_func_name: Union[str, _Unset] = _unset,
        pgid: str,
        declare: Union[str, _Unset] = _unset,
        func: str,
        table: str,
        constraint: Union[str, _Unset] = _unset,
        when: Union["pgtrigger.core.When", str, None],
        operation: Union["pgtrigger.core.Operation", str, None],
        timing: Union[str, _Unset] = _unset,
        referencing: Union["pgtrigger.core.Referencing", str, _Unset] = _unset,
        level: Union["pgtrigger.core.Level", str, _Unset] = _unset,
        condition: Union[str, _Unset] = _unset,
        execute: Union[str, _Unset] = _unset,
        hash: Optional[str] = None,
    ):
        """Initialize the SQL and store it in the `.data` attribute."""
        self.kwargs = {
            key: str(val)
            for key, val in locals().items()
            if key not in ("self", "hash") and val is not _unset
        }
        self.defaults = self.get_defaults(pgid)
        sql_args = {**self.defaults, **self.kwargs, **{"table": utils.quote(table)}}

        self.hash = (
            hash
            or hashlib.sha1(
                self.get_template().format(**{**sql_args, **{"hash": ""}}).encode()
            ).hexdigest()
        )
        self.data = self.get_template().format(**{**sql_args, **{"hash": self.hash}})
        self.pgid = pgid
        self.table = table

    def deconstruct(self) -> tuple[str, Sequence[Any], dict[str, Any]]:
        """
        Serialize the construction of this class so that it can be used in migrations.
        """
        kwargs = {
            key: val for key, val in self.kwargs.items() if self.defaults.get(key, _unset) != val
        }

        path = f"{self.__class__.__module__}.{self.__class__.__name__}"
        return path, [], {**kwargs, **{"hash": self.hash}}


class _TriggerDdlSql(collections.UserString):
    def get_template(self) -> str:
        raise NotImplementedError

    def __init__(self, *, pgid: str, table: str) -> None:
        """Initialize the SQL and store it in the `.data` attribute."""
        sql_args = {**locals(), **{"table": utils.quote(table)}}

        self.data = self.get_template().format(**sql_args)


class DropTriggerSql(_TriggerDdlSql):
    """SQL for dropping a trigger

    Triggers are dropped in migrations, so any edits to
    the drop trigger template should be backwards compatible
    or versioned. I.e. older migrations need to always point to
    the SQL here
    """

    def get_template(self) -> str:
        return "DROP TRIGGER IF EXISTS {pgid} ON {table};"


class EnableTriggerSql(_TriggerDdlSql):
    """SQL for enabling a trigger

    We don't currently perform enabling/disabling in migrations,
    so this SQL can be changed without consequences to past
    migrations.
    """

    def get_template(self) -> str:
        return "ALTER TABLE {table} ENABLE TRIGGER {pgid};"


class DisableTriggerSql(_TriggerDdlSql):
    """SQL for disabling a trigger

    We don't currently perform enabling/disabling in migrations,
    so this SQL can be changed without consequences to past
    migrations.
    """

    def get_template(self) -> str:
        return "ALTER TABLE {table} DISABLE TRIGGER {pgid};"


class Trigger:
    """
    A compiled trigger that's added to internal model state of migrations. It consists
    of a name and the trigger SQL for migrations.
    """

    def __init__(self, *, name: Optional[str], sql: UpsertTriggerSql) -> None:
        self.name = name
        self.sql = sql
        assert isinstance(sql, UpsertTriggerSql)

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__ and self.name == other.name and self.sql == other.sql
        )

    @property
    def install_sql(self) -> str:
        return str(self.sql)

    @property
    def uninstall_sql(self) -> str:
        return str(DropTriggerSql(pgid=self.sql.pgid, table=self.sql.table))

    @property
    def enable_sql(self) -> str:
        return str(EnableTriggerSql(pgid=self.sql.pgid, table=self.sql.table))

    @property
    def disable_sql(self) -> str:
        return str(DisableTriggerSql(pgid=self.sql.pgid, table=self.sql.table))

    @property
    def hash(self) -> str:
        return self.sql.hash

    def deconstruct(self) -> tuple[str, Sequence[Any], dict[str, Any]]:
        """
        Serialize the construction of this class so that it can be used in migrations.
        """
        path = f"{self.__class__.__module__}.{self.__class__.__name__}"
        return path, [], {"name": self.name, "sql": self.sql}



from settings import *

# Turn on pgtrigger migrations in the test suite
PGTRIGGER_MIGRATIONS = True

# We turn this on in tests to ensure that triggers are installed
# when the test database is set up. We dynamically turn it off
# when testing migrations.
PGTRIGGER_INSTALL_ON_MIGRATE = True



# Generated by Django 3.2.15 on 2022-08-11 11:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0008_searchmodel"),
    ]

    operations = [
        migrations.CreateModel(
            name="OrderSchema",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("int_field", models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="ReceiptSchema",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("char_field", models.CharField(max_length=128)),
            ],
            options={
                "db_table": "table.with.dots",
            },
        ),
    ]



# flake8: noqa
from unittest import mock

from django.core.management import call_command
import pytest

import pgtrigger
from pgtrigger import registry


@pytest.mark.django_db
def test_full_ls(capsys):
    """Tests listing all triggers"""

    call_command("pgtrigger", "ls")

    lines = capsys.readouterr().out.split("\n")
    expected_lines = [
        "",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.CustomSoftDelete:soft_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.CustomTableName:protect_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.FSM:fsm",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SearchModel:add_body_title_to_vector",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SearchModel:add_body_to_vector",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestDefaultThrough:protect_it",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTrigger:protect_misc_insert",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.ToLogModel:after_update_row_test",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.ToLogModel:after_update_statement_test",
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.ToLogModel:update_of_statement_test",
    ]
    assert set(expected_lines).issubset(set(lines))


@pytest.mark.django_db
def test_subset_ls(capsys):
    """Tests listing some triggers"""

    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
        ]
    )


@pytest.mark.django_db
def test_main_commands(capsys):
    """
    Tests running main commands
    """

    call_command("pgtrigger", "uninstall")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "",
            "\x1b[91mUNINSTALLED\x1b[0m \x1b[94mN/A\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[91mUNINSTALLED\x1b[0m \x1b[94mN/A\x1b[0m tests.TestTriggerProxy:protect_delete",
        ]
    )

    call_command("pgtrigger", "install")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
        ]
    )

    call_command("pgtrigger", "disable")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "",
            "\x1b[92mINSTALLED\x1b[0m \x1b[91mDISABLED\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m \x1b[91mDISABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
        ]
    )

    call_command("pgtrigger", "enable")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
        ]
    )


@pytest.mark.django_db
def test_prune(capsys):
    """Test pruning a trigger"""
    # Make it appear as though the trigger has been renamed and is no
    # longer installed
    soft_delete_model, soft_delete_trigger = pgtrigger.registered("tests.SoftDelete:soft_delete")[
        0
    ]
    with soft_delete_trigger.unregister(soft_delete_model):
        call_command("pgtrigger", "ls")
        captured = capsys.readouterr()
        lines = sorted(captured.out.split("\n"))
        assert (
            "\x1b[96mPRUNE\x1b[0m     \x1b[92mENABLED\x1b[0m tests_softdelete:pgtrigger_soft_delete_f41be"
        ) in lines

        call_command("pgtrigger", "prune")

    call_command("pgtrigger", "ls")
    lines = capsys.readouterr().out.split("\n")
    assert (
        "\x1b[91mUNINSTALLED\x1b[0m \x1b[94mN/A\x1b[0m     tests.SoftDelete:soft_delete"
    ) in lines

    call_command("pgtrigger", "install")
    call_command("pgtrigger", "ls")
    lines = capsys.readouterr().out.split("\n")
    assert (
        "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete"
    ) in lines


@pytest.mark.django_db(databases=["default", "other"])
def test_outdated(capsys, mocker):
    """Test an outdated trigger"""
    # Make it appear like the trigger is out of date by changing
    # its hash
    mocker.patch.object(
        registry._registry["tests.SoftDelete:soft_delete"][1],
        "compile",
        return_value=mocker.Mock(hash="hash"),
    )

    call_command("pgtrigger", "ls")
    lines = capsys.readouterr().out.split("\n")
    assert (
        "\x1b[93mOUTDATED\x1b[0m  \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete"
    ) in lines


@pytest.mark.django_db
def test_main_commands_w_args(capsys):
    """
    Tests running main commands with arguments
    """

    call_command("pgtrigger", "uninstall", "tests.SoftDelete:soft_delete")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "\x1b[91mUNINSTALLED\x1b[0m \x1b[94mN/A\x1b[0m     tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m   \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
            "",
        ]
    )

    call_command("pgtrigger", "install", "tests.SoftDelete:soft_delete")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
            "",
        ]
    )

    call_command("pgtrigger", "disable", "tests.SoftDelete:soft_delete")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "\x1b[92mINSTALLED\x1b[0m \x1b[91mDISABLED\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m  tests.TestTriggerProxy:protect_delete",
            "",
        ]
    )

    call_command("pgtrigger", "enable", "tests.SoftDelete:soft_delete")
    call_command(
        "pgtrigger",
        "ls",
        "tests.SoftDelete:soft_delete",
        "tests.TestTriggerProxy:protect_delete",
    )

    lines = capsys.readouterr().out.split("\n")
    assert set(lines) == set(
        [
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.SoftDelete:soft_delete",
            "\x1b[92mINSTALLED\x1b[0m \x1b[92mENABLED\x1b[0m tests.TestTriggerProxy:protect_delete",
            "",
        ]
    )
